/////////////////////////////////////////////////////////////
// CINEMA 4D SDK                                           //
/////////////////////////////////////////////////////////////
// (c) MAXON Computer GmbH, all rights reserved            //
/////////////////////////////////////////////////////////////

//Base Effector class
#ifndef C4D_BASEEFFECTORDATA_H__
#define C4D_BASEEFFECTORDATA_H__

#ifdef __API_INTERN__
	#error "Not in the kernel"
#endif

#include "lib_description.h"
#include "c4d_thread.h"
#include "ge_dynamicarray.h"
#include "c4d_objectdata.h"
#include "c4d_basedocument.h"
#include "c4d_baseobject.h"

/// @addtogroup group_effector_id Effector ID
/// @ingroup group_enumeration
/// @{
#define Obaseeffector	 		1018560	///< Base effector object ID.
#define Obasemogen				1018639	///< Base mogen object ID.
#define ID_BASE_EDEFORMER	1019305	///< Base effector deformer ID.
#define ID_MOTAGDATA			1018625	///< ID of the tag containing the MoData on a MoGraph object.
#define ID_MOBAKETAG			1019337	///< MoGraph bake tag ID.
#define Tmgselection			1021338	///< MoGraph selection tag ID.
/// @}

/// @addtogroup group_effector_libid Effector Library ID
/// @ingroup group_enumeration
/// @{
#define BASEEFFECTORLIB_ID 1019554 ///< Base effector library ID.
#define MODATALIB_ID			 440000059 ///< MoGraph data library ID.
/// @}

/// @addtogroup group_msg_effector MSG Effector
/// @ingroup group_enumeration
/// @{
#define MSG_EXECUTE_EFFECTOR			1018630 ///< Sent to execute the effector. The corresponding data is Effector_PassData.
#define MSG_SAMPLE_EFFECTOR_VALUE	1019294 ///< Sent to sample the effector's strength for any particular point in space. The corresponding data is Effector_PassValueData.
#define MSG_GET_MODATA						1019522 ///< Sent to get the MoData. Sent to the MoData tag (ID is ::ID_MOTAGDATA) of the MoGraph object to get the MoData for. The corresponding data is GetMoDataMessage.
#define MSG_GET_MODATASELECTION		1021339 ///< Send to retrieve the data of MoGraph selection tags. The corresponding data is GetMGSelectionMessage.

#define BC_ID_MODATAINDEX					440000062 ///< The index of the clone corresponding to its MoData particle index.
#define BC_ID_MODATATAGINDEX			440000063 ///< The MoData tag index. Set in the clones container.

#define MSG_DISABLE_DELAY_EFFECTOR 440000146 ///< @markPrivate
#define MSG_ENABLE_DELAY_EFFECTOR	 440000147 ///< @markPrivate
#define MSG_MOGRAPH_REEVALUATE		 440000197
/// @}

#define BLEND_COUNT 21 ///< The number of values that a value driven effector has to calculate; normally hard coded (do not change).

/// @addtogroup EFFECTORFLAGS
/// @ingroup group_enumeration
/// @{
/// Effector flags.
/// @see EffectorData::GetEffectorFlags()
#define EFFECTORFLAGS_HASFALLOFF			(1 << 0) ///< A falloff control panel will be added to the effector and a falloff will be available to sample.
#define EFFECTORFLAGS_TIMEDEPENDENT		(1 << 1) ///< The effector will be forced to update with changes to the time.
#define EFFECTORFLAGS_CAMERADEPENDENT	(1 << 2) ///< The effector will always be dirty and forced to update.
/// @}

/// @addtogroup MOGENFLAG
/// @ingroup group_enumeration
/// @{
/// MoData flags.
/// @see ::MODATA_FLAGS
#define MOGENFLAG_CLONE_ON	(1 << 0)	///< Particle is visible.
#define MOGENFLAG_DISABLE		(1 << 1)	///< Particle is permanently disabled.
#define MOGENFLAG_BORN			(1 << 2)	///< Particle is just generated. @markInternal
#define MOGENFLAG_MODATASET	(1 << 3)	///< The MoData has been set and does not need the input of the transform panel.
#define MOGENFLAG_COLORSET	(1 << 4)	///< The MoData color has been set and does not need to be updated.
#define MOGENFLAG_TIMESET		(1 << 5)	///< The MoData time has been set and does not need to be updated.
/// @}

/// @addtogroup MDDIRTY
/// @ingroup group_enumeration
/// @{
/// MoData dirty flags.
#define MDDIRTY_ARRAYCOUNT (1 << 0)		///< Array count itself changed (number of different arrays rather than length of arrays).
#define MDDIRTY_COUNT			 (1 << 1)		///< Length of the arrays changed.
#define MDDIRTY_DATA			 (1 << 2)		///< Data in the arrays changed, must be manually set.
/// @}

class C4D_Falloff;

/// @addtogroup MD_TYPE
/// @ingroup group_enumeration
/// @{
enum MD_TYPE
{
	MD_NONE		=	DA_NIL,				///< None.
	MD_CHAR		= 40000000,			///< ::Char type.
	MD_UCHAR	=	40000001,			///< ::UChar type.
	MD_LONG		= DTYPE_LONG,		///< ::Int32 type.
	MD_ULONG	=	40000002,			///< ::UInt32 type.
	MD_LLONG	=	DA_LLONG,			///< ::Int64 type.
	MD_MATRIX	= DA_MATRIX,		///< ::Matrix type.
	MD_COLOR	= DTYPE_COLOR,	///< Color type (::Vector).
	MD_VECTOR	=	DTYPE_VECTOR,	///< ::Vector type.
	MD_NORMAL	= DTYPE_NORMAL,	///< Normal type (::Vector).
	MD_REAL		= DTYPE_REAL,		///< ::Float type.
	MD_BOOL		= DTYPE_BOOL		///< ::Bool type.
} ENUM_END_LIST(MD_TYPE);
/// @}

/// @addtogroup MODATA_IDS
/// @ingroup group_enumeration
/// @{
/// MoData arrays ID.
enum
{
	MODATA_MATRIX	=	40000000,	///< ::Matrix. Matrix of the clone.
	MODATA_COLOR,							///< ::Vector. Color of the clone.
	MODATA_SIZE,							///< ::Vector. Size of the clone.
	MODATA_UVW,								///< ::Vector. UV position of the clone.
	MODATA_FLAGS,							///< ::Int32. MoData flags: @enumerateEnum{@ref MOGENFLAG}
	MODATA_WEIGHT,						///< ::Float. Weight of the clone.
	MODATA_CLONE,							///< ::Float. Clone offset (picks which child of the Cloner gets cloned or the blending between those children).
	MODATA_TIME,							///< ::Float. Time offset of the clone.
	MODATA_LASTMAT,						///< ::Matrix. Previous frame particle matrix.
	MODATA_STARTMAT,					///< ::Matrix. Matrix at the particle's birth.
	MODATA_ALT_INDEX,					///< ::Int32. Alternative index value that can be used for instance by the Step Effector when cloned over a spline with an offset.
	MODATA_FALLOFF_WGT,				///< ::Float. Falloff weight.
	MODATA_SPLINE_SEGMENT,		///< ::Int32. Segment index, mostly used with the MoSpline (currently unused).
	MODATA_GROWTH							///< ::Float. Offset of growth for the particle on the MoSpline (currently unused).
};
/// @}


//----------------------------------------------------------------------------------------
/// Get the default value for the specified MoData array @formatParam{id}.
///
/// @param[in] id									The MoData array ID: @enumerateEnum{@ref MODATA_IDS}
/// @return												The MoData default value.
//----------------------------------------------------------------------------------------
inline GeData GetMoDataDefault(Int32 id)
{
	switch (id)
	{
		case MODATA_LASTMAT:
		case MODATA_STARTMAT:
		case MODATA_MATRIX: return GeData(Matrix()); break;
		case MODATA_COLOR: return GeData(Vector(0.5)); break;
		case MODATA_SIZE: return GeData(Vector(1.0)); break;
		case MODATA_UVW: return GeData(Vector(0.0)); break;
		case MODATA_FLAGS: return GeData(MOGENFLAG_CLONE_ON | MOGENFLAG_BORN); break;
		case MODATA_WEIGHT:
		case MODATA_CLONE:
		case MODATA_TIME: return GeData(0.0); break;
		case MODATA_SPLINE_SEGMENT: return GeData(0); break;
		case MODATA_ALT_INDEX: return GeData(NOTOK); break;
		case MODATA_FALLOFF_WGT:
		case MODATA_GROWTH: return GeData(1.0); break;
		default: break;
	}

	return GeData();
};

//----------------------------------------------------------------------------------------
/// Get the default type for the specified MoData array @formatParam{id}.
///
/// @param[in] id									The MoData array ID: @enumerateEnum{@ref MODATA_IDS}
/// @return												The default type: @enumerateEnum{::MD_TYPE}
//----------------------------------------------------------------------------------------
inline MD_TYPE GetMoDataDefaultType(Int32 id)
{
	switch (id)
	{
		case MODATA_LASTMAT:
		case MODATA_STARTMAT:
		case MODATA_MATRIX: return MD_MATRIX; break;
		case MODATA_COLOR: return MD_COLOR; break;
		case MODATA_SIZE:
		case MODATA_UVW: return MD_VECTOR; break;
		case MODATA_SPLINE_SEGMENT:
		case MODATA_ALT_INDEX:
		case MODATA_FLAGS: return MD_LONG; break;
		case MODATA_WEIGHT:
		case MODATA_CLONE:
		case MODATA_FALLOFF_WGT:
		case MODATA_GROWTH:
		case MODATA_TIME: return MD_REAL; break;
		default: break;
	}

	return MD_NONE;
};

/// @addtogroup MDARRAYFLAG
/// @ingroup group_enumeration
/// @{
#define MDARRAYFLAG_NOTSEENBYEFFECTOR	 (1 << 0)	///< Data is not seen and modified by the effector.
#define MDARRAYFLAG_SEPERATECOMPONENTS (1 << 1)	///< Vector should have separate components in effector (e.g. UV).
#define MDARRAYFLAG_2DVECTOR					 (1 << 2)	///< Vector is a @em 2D vector rather than @em 3D (e.g. UV).
/// @}

/// @addtogroup ID_MODATA_PASS
/// @ingroup group_enumeration
/// @{
enum
{
/// This is used to allow the delay effector to work with the MoGraph extrude deformer, each level of extrusion gets a pass index which is set in the MoData's own container (MoData::GetDataInstance()):
/// @code
/// Int32 pass = md->GetDataInstance()->GetInt32(ID_MODATA_PASS, 0);
/// @endcode
/// By doing this the delay effector can differentiate the levels and create a reasonable effect.\n
/// The idea is simply that there can be multiple passes through the effectors within a single execution.\n
/// And if ::ID_MODATA_PASS is set to a pass number then the delay effector (and anything else that wants to use it) should work.
	ID_MODATA_PASS = 100
};
/// @}


//-------------------------------------------------------------------------------------------------
/// Motion particle transformation data.
struct Dpoint
{
	Matrix m; ///< Particle raw matrix.

	//----------------------------------------------------------------------------------------
	/// Get the particle position.
	///
	/// @return												The particle position.
	//----------------------------------------------------------------------------------------
	inline Vector GetPos(void) const
	{
		return m.off;
	}

	//----------------------------------------------------------------------------------------
	/// Get the particle size.
	///
	/// @return												The particle size.
	//----------------------------------------------------------------------------------------
	inline Vector GetSize(void)
	{
		return Vector(Len(m.v1), Len(m.v2), Len(m.v3));
	}

	//----------------------------------------------------------------------------------------
	/// Get the particle matrix.
	///
	/// @param[out] h									Assigned the particle matrix.
	/// @param[out] size							Assigned the particle size.
	//----------------------------------------------------------------------------------------
	inline void GetMatrix(Matrix& h, const Vector& size) const
	{
		h = m;
		h.v1 -= h.off;
		h.v2 -= h.off;
		h.v3 -= h.off;
		h.v1	= !Cross(h.v2, h.v3) * size.x;
		h.v2	= !Cross(h.v3, h.v1) * size.y;
		h.v3	= !Cross(h.v1, h.v2) * size.z;
	}
};


//-------------------------------------------------------------------------------------------------
/// Struct that holds the information for each particle.
struct Ipoint
{
public:
	Ipoint()
	{
		flags = MOGENFLAG_CLONE_ON | MOGENFLAG_BORN;
		color = Vector(0.5);
		size = Vector(1.0);
		weight = 0.0;
		clone = 1.0;
		uvw = Vector(0.0);
		time = 0.0;
		alt_index = NOTOK;
	}
	~Ipoint()
	{
	}

	Vector color;			///< Color of the particle.
	Vector size;			///< Size of the particle.
	Vector uvw;				///< Particle direct UVW position.
	Int32 flags;			///< Particle flags: @enumerateEnum{@ref MOGENFLAG}
	Float weight;			///< Offset against falloff value to allow effectors to drive other effectors.
	Float clone;			///< Index of the child that will be cloned in the cloner object.
	Float time;				///< Delta time for each clone.
	Matrix lastmat;		///< Last position.
	Matrix startmat;	///< Starting position.
	Int32 alt_index;	///< Alternative index for particles, instances etc. with some effectors, e.g. step.
};


//-------------------------------------------------------------------------------------------------
/// Template array class used for arrays filled by MoData.
template <class TYPE> class MDArray
{
private:
	TYPE* ptr;
	mutable TYPE dv;
	Int count;

public:
	//-------------------------------------------------------------------------------------------------
	/// Default constructor.
	//-------------------------------------------------------------------------------------------------
	MDArray(void) { ptr = nullptr; dv = TYPE(); count = 0; }

	//----------------------------------------------------------------------------------------
	/// Constructor with the array pointer and the default value.
	///
	/// @param[in] array_pointer			The array pointer.
	/// @param[in] array_count				The number of elements in @formatParam{array_pointer}.
	/// @param[in] default_value			The default value.
	//----------------------------------------------------------------------------------------
	MDArray(TYPE* array_pointer, Int32 array_count, TYPE default_value/*= TYPE()*/)
	{
		ptr = array_pointer;
		dv = default_value;
		count = array_count;
	}

	//-------------------------------------------------------------------------------------------------
	/// Default destructor.
	//-------------------------------------------------------------------------------------------------
	~MDArray(void) { }

	//----------------------------------------------------------------------------------------
	/// Fills the elements of the array with @formatParam{default_value}.
	///
	/// @param[in] default_value			The value to fill with.
	//----------------------------------------------------------------------------------------
	void Fill(const TYPE& default_value)
	{
		if (ptr && count)
		{
			Int i = 0;
			for (i = count - 1; i >= 0; --i)
				ptr[i] = default_value;
		}
		dv = default_value;
	}

	//-------------------------------------------------------------------------------------------------
	/// Conversion operator.
	//-------------------------------------------------------------------------------------------------
	operator TYPE* ()
	{
		return ptr;
	}

	//----------------------------------------------------------------------------------------
	/// Get the array pointer.
	///
	/// @return												The array pointer.
	//----------------------------------------------------------------------------------------
	TYPE* GetPointer(void) { return ptr; }

	//----------------------------------------------------------------------------------------
	/// Check if the array contains any elements.
	///
	/// @return												@trueIfOtherwiseFalse{the array is empty}
	//----------------------------------------------------------------------------------------
	Bool operator!() { return ptr == nullptr; }

	//----------------------------------------------------------------------------------------
	/// Array access operator.
	///
	/// @param[in] i									The element index in the array.
	/// @return												The element at @formatParam{i} in the array.
	//----------------------------------------------------------------------------------------
	TYPE& operator[] (Int32 i) const { DebugAssert(!ptr || (i >= 0 && i < count)); return (!ptr || i < 0 || i >= count) ? dv : ptr[i]; }

	//----------------------------------------------------------------------------------------
	/// Array access operator.
	///
	/// @param[in] i									The element index in the array.
	/// @return												The element at @formatParam{i} in the array.
	//----------------------------------------------------------------------------------------
	TYPE& operator[] (Int64 i) const { DebugAssert(!ptr || (i >= 0 && i < count)); return (!ptr || i < 0 || i >= count) ? dv : ptr[i]; }

	//----------------------------------------------------------------------------------------
	/// Retrieve a const pointer to the internal pointer.
	///
	/// @return												The internal pointer.
	//----------------------------------------------------------------------------------------
	TYPE* const* operator &() { return &ptr; }
};


//-------------------------------------------------------------------------------------------------
/// Structure for passing cloned points around.\n
/// The data is stored in separate dynamic arrays for internal purposes in the MoData struct. The Matrices are always in local space to the object that generated/owns the particles.
///
/// @see MoData::SetWithEx
/// @warning Only kept for legacy support. Has been replaced by the MoData class.
//-------------------------------------------------------------------------------------------------
struct MoDataEx
{
	Int32 count; ///< The number of particles. @warning Should match the length of both @link mat@endlink and @link data@endlink arrays.
	Int32 pass;  ///<  The pass for multi-index calls.
	GeDynamicArray<Dpoint> mat;  ///< The array of matrices.
	GeDynamicArray<Ipoint> data; ///< The array for the other data.
};


//-------------------------------------------------------------------------------------------------
/// Class for handling MoGraph motion data.
//-------------------------------------------------------------------------------------------------
class MoData
{
private:
	MoData(void);
	~MoData(void);

	MoData& operator = (const MoData& d);
	MoData* operator = (MoData* d);

public:
	//----------------------------------------------------------------------------------------
	/// @allocatesA{motion data}
	///
	/// @return												@allocReturn{motion data}
	//----------------------------------------------------------------------------------------
	static MoData* Alloc(void);

	//----------------------------------------------------------------------------------------
	/// @destructsAlloc{motion data}
	///
	/// @param[in,out] d							@theToDestruct{motion data}
	//----------------------------------------------------------------------------------------
	static void Free(MoData*& d);

	/// @name Dirty
	/// @{

	//----------------------------------------------------------------------------------------
	/// Get the dirty count. Can be used to check if something has changed.
	///
	/// @param[in] mask								The dirty flag mask. Can be a combination of: @enumerateEnum{::DIRTYFLAGS}
	/// @return												The dirty count.
	//----------------------------------------------------------------------------------------
	UInt32 GetDirty(DIRTYFLAGS mask = DIRTYFLAGS_0);

	//----------------------------------------------------------------------------------------
	/// Mark the data as dirty.
	///
	/// @param[in] mask								The dirty flag mask. Can be a combination of: @enumerateEnum{::DIRTYFLAGS}
	//----------------------------------------------------------------------------------------
	void SetDirty(DIRTYFLAGS mask = DIRTYFLAGS_0);

	/// @}

	//----------------------------------------------------------------------------------------
	/// Initialize the data with MoDataEx.
	///
	/// @param[in] md									The data to use for initializing.
	/// @return												@trueIfOtherwiseFalse{motion data was successfully initialized}
	//----------------------------------------------------------------------------------------
	Bool SetWithEx(const MoDataEx& md);

	//----------------------------------------------------------------------------------------
	/// Clear the data in the arrays.
	///
	/// @param[in] reset							If @formatBuiltInConstant{true}, the arrays will be filled with their default values.
	//----------------------------------------------------------------------------------------
	void Clear(Bool reset);

	/// @name Read/Write
	/// @{

	//----------------------------------------------------------------------------------------
	/// Read the data from a hyper file.
	///
	/// @param[in] hf									The hyper file to read from. @callerOwnsPointed{hyper file}
	/// @return												@trueIfOtherwiseFalse{the data was read}
	//----------------------------------------------------------------------------------------
	Bool Read(HyperFile* hf);

	//----------------------------------------------------------------------------------------
	/// Write the data to a hyper file.
	///
	/// @param[in] hf									The hyper file to write to. @callerOwnsPointed{hyper file}
	/// @return												@trueIfOtherwiseFalse{the data was written}
	//----------------------------------------------------------------------------------------
	Bool Write(HyperFile* hf);

	/// @}

	//----------------------------------------------------------------------------------------
	/// Get the size of the data in bytes.
	///
	/// @return												The size of the data.
	//----------------------------------------------------------------------------------------
	Int GetMemorySize(void);

	/// @name Arrays Length & Number
	/// @{

	//----------------------------------------------------------------------------------------
	/// Set the length of the arrays.
	///
	/// @param[in] cnt								The new length of the arrays.
	/// @return												@trueIfOtherwiseFalse{the length was set}
	//----------------------------------------------------------------------------------------
	Bool SetCount(Int cnt);

	//----------------------------------------------------------------------------------------
	/// Get the length of the arrays.
	///
	/// @return												The length of the arrays.
	//----------------------------------------------------------------------------------------
	Int GetCount(void);

	//----------------------------------------------------------------------------------------
	/// Get the number of arrays.
	///
	/// @return												The number of arrays.
	//----------------------------------------------------------------------------------------
	Int32 GetArrayCount(void);

	/// @}

	/// @name Array Information
	/// @{

	//----------------------------------------------------------------------------------------
	/// Get the description ID for the specified array @formatParam{index}.
	///
	/// @param[in] index							The array index. @em 0 <= @formatParam{index} < GetArrayCount()
	/// @return												The description ID.
	//----------------------------------------------------------------------------------------
	DescID GetArrayDescID(Int32 index) const;

	//----------------------------------------------------------------------------------------
	/// Get the ID for the specified array @formatParam{index}.
	///
	/// @param[in] index							The array index. @em 0 <= @formatParam{index} < GetArrayCount()
	/// @return												The array ID: @enumerateEnum{@ref MODATA_IDS}
	//----------------------------------------------------------------------------------------
	Int32 GetArrayID(Int32 index) const;

	//----------------------------------------------------------------------------------------
	/// Get the data type of the specified array @formatParam{index}.
	///
	/// @param[in] index							The array index. @em 0 <= @formatParam{index} < GetArrayCount()
	/// @return												The data type: @enumerateEnum{@ref DTYPE}
	//----------------------------------------------------------------------------------------
	Int32 GetArrayIndexType(Int32 index);

	//----------------------------------------------------------------------------------------
	/// Get the data type of the specified array @formatParam{id}.
	///
	/// @param[in] id									The array ID: @enumerateEnum{@ref MODATA_IDS}
	/// @return												The type: @enumerateEnum{@ref DTYPE}
	//----------------------------------------------------------------------------------------
	Int32 GetArrayType(Int32 id);

	//----------------------------------------------------------------------------------------
	/// Get the array index for the specified description @formatParam{id}.
	///
	/// @param[in] id									The description ID.
	/// @return												The array index.
	//----------------------------------------------------------------------------------------
	Int GetArrayIndex(const DescID& id);

	//----------------------------------------------------------------------------------------
	/// Get the array index for the specified array @formatParam{id}.
	///
	/// @param[in] id									The array ID: @enumerateEnum{@ref MODATA_IDS}
	/// @return												The  array index.
	//----------------------------------------------------------------------------------------
	Int GetArrayIndex(Int32 id);

	/// @}

	/// @name Get/Set Data Container
	/// @{

	//----------------------------------------------------------------------------------------
	/// Get a pointer to the container for the specified array description @formatParam{id}.
	///
	/// @param[in] id									The description ID.
	/// @return												A pointer to the internal container. @theOwnsThePointed{motion data object, container}
	//----------------------------------------------------------------------------------------
	BaseContainer* GetDataInstance(const DescID& id);

	//----------------------------------------------------------------------------------------
	/// Get a pointer to the container for the specified array @formatParam{id}.
	///
	/// @param[in] id									The array ID: @enumerateEnum{@ref MODATA_IDS}
	/// @return												A pointer to the internal container. @theOwnsThePointed{motion data object, container}
	//----------------------------------------------------------------------------------------
	BaseContainer* GetDataInstance(Int32 id = NOTOK);

	//----------------------------------------------------------------------------------------
	/// Get a pointer to the container for the specified array @formatParam{index}.
	///
	/// @param[in] index							The array index. @em 0 <= @formatParam{index} < GetArrayCount()
	/// @return												A pointer to the internal container. @theOwnsThePointed{motion data object, container}
	//----------------------------------------------------------------------------------------
	BaseContainer* GetDataIndexInstance(Int32 index);

	//----------------------------------------------------------------------------------------
	/// Get a copy of the container for the specified array @formatParam{id}.
	///
	/// @param[in] id									The array ID: @enumerateEnum{@ref MODATA_IDS}
	/// @return												A copy of the array's container.
	//----------------------------------------------------------------------------------------
	BaseContainer GetData(Int32 id = NOTOK);

	//----------------------------------------------------------------------------------------
	/// Set the the container for the specified array @formatParam{id}.
	///
	/// @param[in] bc									The container to set.
	/// @param[in] id									The array ID: @enumerateEnum{@ref MODATA_IDS}
	//----------------------------------------------------------------------------------------
	void SetData(const BaseContainer& bc, Int32 id = NOTOK);

	/// @}

	//----------------------------------------------------------------------------------------
	/// Must be used to lock the class with AutoLocker prior to accessing the MoData values directly to prevent threading issues and therefore crashes.
	/// @code
	/// AutoLocker al(modata.GetAutoLock());
	/// @endcode
	///
	/// @return												The AutoLock used for locking the class.
	//----------------------------------------------------------------------------------------
	AutoLock& GetAutoLock(void);

	/// @name Add/Remove Array
	/// @{

	//----------------------------------------------------------------------------------------
	/// Add the specified array.
	///
	/// @param[in] id									The description ID of the array.
	/// @param[in] name								The name of the array.
	/// @param[in] default_flags			The default flags: @enumerateEnum{@ref MOGENFLAG}
	/// @return												The index of the added array or ::NOTOK if the method failed.
	//----------------------------------------------------------------------------------------
	Int32 AddArray(const DescID& id, String name = "", Int32 default_flags = 0);

	//----------------------------------------------------------------------------------------
	/// Add the specifed array.
	///
	/// @param[in] id									The array ID: @enumerateEnum{@ref MODATA_IDS}
	/// @param[in] type								The type: @enumerateEnum{@ref DTYPE}
	/// @param[in] name								The name of the array.
	/// @param[in] default_flags			The default flags: @enumerateEnum{@ref MOGENFLAG}
	/// @return												The index of the added array or ::NOTOK if the method failed.
	//----------------------------------------------------------------------------------------
	Int32 AddArray(Int32 id, Int32 type, String name = "", Int32 default_flags = 0);

	//----------------------------------------------------------------------------------------
	/// Remove the specified array.
	///
	/// @param[in] id									The description ID of the array.
	/// @return												@trueIfOtherwiseFalse{the array was removed}
	//----------------------------------------------------------------------------------------
	Bool RemoveArray(const DescID& id);

	//----------------------------------------------------------------------------------------
	/// Remove the specified array.
	///
	/// @param[in] id									The array ID: @enumerateEnum{@ref MODATA_IDS}
	/// @return												@trueIfOtherwiseFalse{the array was removed}
	//----------------------------------------------------------------------------------------
	Bool RemoveArray(Int32 id);

	/// @}

	/// @name Array Flags
	/// @{

	//-------------------------------------------------------------------------------------------------
	/// @markUnused
	//-------------------------------------------------------------------------------------------------
	Int32 GetFlags(const DescID& id);

	//-------------------------------------------------------------------------------------------------
	/// @markUnused
	//-------------------------------------------------------------------------------------------------
	Int32 GetFlags(Int32 id);

	//-------------------------------------------------------------------------------------------------
	/// @markUnused
	//-------------------------------------------------------------------------------------------------
	Int32 GetIndexFlags(Int32 index);

	//-------------------------------------------------------------------------------------------------
	/// @markUnused
	//-------------------------------------------------------------------------------------------------
	void SetFlags(const DescID& id, Int32 flags);

	//-------------------------------------------------------------------------------------------------
	/// @markUnused
	//-------------------------------------------------------------------------------------------------
	void SetFlags(Int32 id, Int32 flags);

	//-------------------------------------------------------------------------------------------------
	/// @markUnused
	//-------------------------------------------------------------------------------------------------
	void SetIndexFlags(Int32 index, Int32 flags);

	//-------------------------------------------------------------------------------------------------
	/// @markUnused
	//-------------------------------------------------------------------------------------------------
	void SetFlag(const DescID& id, Int32 flag);

	//-------------------------------------------------------------------------------------------------
	/// @markUnused
	//-------------------------------------------------------------------------------------------------
	void SetFlag(Int32 id, Int32 flag);

	//-------------------------------------------------------------------------------------------------
	/// @markUnused
	//-------------------------------------------------------------------------------------------------
	void SetIndexFlag(Int32 index, Int32 flag);

	//-------------------------------------------------------------------------------------------------
	/// @markUnused
	//-------------------------------------------------------------------------------------------------
	void DelFlag(const DescID& id, Int32 flag);

	//-------------------------------------------------------------------------------------------------
	/// @markUnused
	//-------------------------------------------------------------------------------------------------
	void DelFlag(Int32 id, Int32 flag);

	//-------------------------------------------------------------------------------------------------
	/// @markUnused
	//-------------------------------------------------------------------------------------------------
	void DelIndexFlag(Int32 index, Int32 flag);

	/// @}

	/// @name Array name
	/// @{

	//----------------------------------------------------------------------------------------
	/// Set the name for the specified array description @formatParam{id}.
	///
	/// @param[in] id									The description ID of the array.
	/// @param[in] name								The name of the array to set.
	//----------------------------------------------------------------------------------------
	void SetName(const DescID& id, const String& name);

	//----------------------------------------------------------------------------------------
	/// Set the name for the specified array @formatParam{id}.
	///
	/// @param[in] id									The array ID: @enumerateEnum{@ref MODATA_IDS}
	/// @param[in] name								The name of the array to set.
	//----------------------------------------------------------------------------------------
	void SetName(Int32 id, const String& name);

	//----------------------------------------------------------------------------------------
	/// Set the name for the specified array @formatParam{index}.
	///
	/// @param[in] index							The array index. @em 0 <= @formatParam{index} < GetArrayCount()
	/// @param[in] name								The name of the array to set.
	//----------------------------------------------------------------------------------------
	void SetIndexName(Int32 index, const String& name);

	//----------------------------------------------------------------------------------------
	/// Get the name of the specified array description @formatParam{id}.
	///
	/// @param[in] id									The description ID of the array.
	/// @return												The name of the array.
	//----------------------------------------------------------------------------------------
	String GetName(const DescID& id);

	//----------------------------------------------------------------------------------------
	/// Get the name of the specified array @formatParam{id}.
	///
	/// @param[in] id									The array ID: @enumerateEnum{@ref MODATA_IDS}
	/// @return												The name of the array.
	//----------------------------------------------------------------------------------------
	String GetName(Int32 id);

	//----------------------------------------------------------------------------------------
	/// Get the name of the specified array @formatParam{index}.
	///
	/// @param[in] index							The array index. @em 0 <= @formatParam{index} < GetArrayCount()
	/// @return												The name of the array.
	//----------------------------------------------------------------------------------------
	String GetIndexName(Int32 index);

	/// @}

	//Get arrays using DescIDs
	void* GetArray(const DescID& id) const;
	MDArray<Char> GetCharArray(const DescID& id, Char default_value = 0) const;
	MDArray<UChar> GetUCharArray(const DescID& id, UChar default_value = 0) const;
	MDArray<Int32> GetLongArray(const DescID& id, Int32 default_value = 0) const;
	MDArray<UInt32> GetULongArray(const DescID& id, UInt32 default_value = 0) const;
	MDArray<Bool> GetBoolArray(const DescID& id, Bool default_value = true) const;
	MDArray<Int64> GetLLongArray(const DescID& id, Int64 default_value = 0) const;
	MDArray<Float> GetRealArray(const DescID& id, Float default_value = 1.0) const;
	MDArray<Matrix> GetMatrixArray(const DescID& id, Matrix default_value = Matrix()) const;
	MDArray<Vector> GetVectorArray(const DescID& id, Vector default_value = Vector()) const;

	//Get arrays using Int32 ids
	void* GetArray(Int32 id, MD_TYPE type = MD_NONE) const;
	MDArray<Char> GetCharArray(Int32 id, Char default_value = 0) const;
	MDArray<UChar> GetUCharArray(Int32 id, UChar default_value = 0) const;
	MDArray<Int32> GetLongArray(Int32 id, Int32 default_value = 0) const;
	MDArray<UInt32> GetULongArray(Int32 id, UInt32 default_value = 0) const;
	MDArray<Bool> GetBoolArray(Int32 id, Bool default_value = true) const;
	MDArray<Int64> GetLLongArray(Int32 id, Int64 default_value = 0) const;
	MDArray<Float> GetRealArray(Int32 id, Float default_value = 1.0) const;
	MDArray<Matrix> GetMatrixArray(Int32 id, Matrix default_value = Matrix()) const;
	MDArray<Vector> GetVectorArray(Int32 id, Vector default_value = Vector()) const;

	//Get arrays using Int32 indexes
	void* GetIndexArray(Int32 index) const;
	MDArray<Char> GetCharIndexArray(Int32 index) const;
	MDArray<UChar> GetUCharIndexArray(Int32 index) const;
	MDArray<Int32> GetLongIndexArray(Int32 index) const;
	MDArray<UInt32> GetULongIndexArray(Int32 index) const;
	MDArray<Bool> GetBoolIndexArray(Int32 index) const;
	MDArray<Int64> GetLLongIndexArray(Int32 index) const;
	MDArray<Float> GetRealIndexArray(Int32 index) const;
	MDArray<Matrix> GetMatrixIndexArray(Int32 index) const;
	MDArray<Vector> GetVectorIndexArray(Int32 index) const;

	Bool CopyTo(MoData* dest, Bool merge_data = false, Int max_count = NOTOK);
	Bool MergeData(MoData* source, Float percent);

	void Flush(void);
	void SetOffset(Int offset = 0);
	void SetLimit(Int limit = NOTOK);
};

// Structure for passing modata around
struct GetMoDataMessage
{
	GetMoDataMessage(void) { index = 0; modata = nullptr; user_owned = false; }
	~GetMoDataMessage(void)
	{
		if (user_owned && modata)
			MoData::Free(modata);
	}
	MoData* modata;				//The returned MoData
	Int32		index;				//The index of the MoData, in a Text Object there can be 4 MoDatas, Paragraph, Line, Word, Letter
	Bool		user_owned;		//The Modata is either owned by C4D or by the User, if it is User owned then the user must free the modata

	MoData* Release(void)	//Release the modata to the user, they must then free the result
	{
		if (!user_owned)
		{
			MoData* md = MoData::Alloc();
			if (!md)
				return nullptr;
			if (!modata->CopyTo(md))
			{
				MoData::Free(md);
				return nullptr;
			}
			modata = md;
		}
		user_owned = false;
		return modata;
	}
};

// Structure for passing mograph selectino data around
struct GetMGSelectionMessage
{
	BaseSelect* sel;
};

//-------------------
// structure to pass with execute effector message
struct Effector_PassData
{
	Effector_PassData (void)
	{
		op = nullptr;
		md = nullptr;
		weight = 1.0;
		thread = nullptr;
	}
	BaseObject* op;
	MoData*			md;
	Float				weight;
	BaseThread* thread;
};

// structure to get pass with sample effector value
struct Effector_PassValueData
{
	Effector_PassValueData (void)
	{
		value = 0.0;
		pos = Vector(0.0);
	}
	Float	 value;
	Vector pos;
};

struct EffectorStrengths
{
	Vector pos, scale, rot, col, other, other2, other3;
};

struct EffectorDataStruct
{
	EffectorDataStruct(void)
	{
		falloff_time = 0.0;
		strength = 1.0;
		maxstrength = 1.0;
		minstrength = 0.0;
		trans_mode	= col_mode = blend_mode = 0;
		ClearMemType(strengths, BLEND_COUNT);
	}
	~EffectorDataStruct(void)
	{
	}
	Float	 falloff_time;
	Vector pos, scale, rot, col;		//position, scale, rotation, color
	Float	 strength;								//effector strength
	Float	 maxstrength;							//effector minimum strength
	Float	 minstrength;							//effector minimum strength
	Int32	 trans_mode;							//transform mode
	Int32	 col_mode;								//color mode
	Int32	 blend_mode;							//blending mode
	Float	 strengths[BLEND_COUNT];	//amount of strengths for each value, of in order, position, scale, rotation, color (anything else we want to affect)
};

struct EffectorDataParameters;

class EffectorData : public ObjectData
{
public:
	EffectorDataParameters* edata;

	//-----------------------------------------------------
	//Inherited from ObjectData
	virtual Bool Init(GeListNode* node);
	virtual void Free(GeListNode* node);
	virtual Bool GetDDescription(GeListNode* node, Description* description, DESCFLAGS_DESC& flags);
	virtual DRAWRESULT Draw(BaseObject* op, DRAWPASS drawpass, BaseDraw* bd, BaseDrawHelp* bh);
	virtual void GetDimension(BaseObject* op, Vector* mp, Vector* rad);
	virtual Bool AddToExecution(BaseObject* op, PriorityList* list);
	virtual EXECUTIONRESULT	Execute(BaseObject* op, BaseDocument* doc, BaseThread* bt, Int32 priority, EXECUTIONFLAGS flags);
	virtual Bool ModifyObject(BaseObject* mod, BaseDocument* doc, BaseObject* op, const Matrix& op_mg, const Matrix& mod_mg, Float lod, Int32 flags, BaseThread* thread);
	virtual Bool Message(GeListNode* node, Int32 type, void* t_data);
	virtual Int32	GetHandleCount(BaseObject* op);
	virtual void GetHandle(BaseObject* op, Int32 i,	HandleInfo& info);
	virtual void SetHandle(BaseObject* op, Int32 i, Vector p, const HandleInfo& info);
	virtual Bool CopyTo(NodeData* dest, GeListNode* snode, GeListNode* dnode, COPYFLAGS flags, AliasTrans* trn);
	virtual Bool IsInstanceOf(const GeListNode* node, Int32 type) const;

	//Internal versions of inherited routines
	virtual Bool InitEffector(GeListNode* node);
	virtual void FreeEffector(GeListNode* node);

	//-----------------------------------------------------
	//Effector Routines
	EffectorDataStruct*						GetEffectorData(void);
	C4D_Falloff*									GetFalloff(void);
	void AddEffectorDependence(BaseObject* op);

	virtual Int32	GetEffectorFlags(void);
	virtual Bool ModifyDDescription(GeListNode* node, Description* description, AtomArray* ar);

	virtual Bool ExecuteEffector(BaseObject* op, BaseDocument* doc, BaseObject* gen, MoData* md, Float strength, BaseThread* thread);
	virtual void InitPoints(BaseObject* op, BaseObject* gen, BaseDocument* doc, EffectorDataStruct* data, MoData* md, BaseThread* thread);
	virtual void ModifyPoints(BaseObject* op, BaseObject* gen, BaseDocument* doc, EffectorDataStruct* data, MoData* md, BaseThread* thread);
	virtual void FreePoints(void);

	//------------------------------------------------------
	//for random etc, where a value must be calculated even if it's not used
	virtual void CalcPlacebo(BaseObject* op, BaseObject* gen, BaseDocument* doc, EffectorDataStruct* data, Int32 index, MoData* md, const Vector& globalpos, Float fall_weight);

	//modify data->strengths values accordingly in CalcPointValue
	virtual void CalcPointValue(BaseObject* op, BaseObject* gen, BaseDocument* doc, EffectorDataStruct* data, Int32 index, MoData* md, const Vector& globalpos, Float fall_weight);

	//called separately as it may not be needed (and therefore sometimes might reduce number of calculations
	virtual Vector CalcPointColor(BaseObject* op, BaseObject* gen, BaseDocument* doc, EffectorDataStruct* data, Int32 index, MoData* md, const Vector& globalpos, Float fall_weight);
};

Bool RegisterEffectorPlugin(Int32 id, const String& str, Int32 info, DataAllocator* g, const String& description, BaseBitmap* icon, Int32 disklevel);

#endif // C4D_BASEEFFECTORDATA_H__

