/////////////////////////////////////////////////////////////
// CINEMA 4D SDK                                           //
/////////////////////////////////////////////////////////////
// (c) MAXON Computer GmbH, all rights reserved            //
/////////////////////////////////////////////////////////////

#ifndef LIB_PAINT_H__
#define LIB_PAINT_H__

#include "c4d_library.h"
#include "c4d_memory.h"

#define C4D_PAINT_LIBRARY_ID	1031402

//----------------------------------------------------------------------------------------
///	Functions to access Texture Views
//----------------------------------------------------------------------------------------
struct PaintManager
{
	//----------------------------------------------------------------------------------------
	///	Get the number of opened Texture Views.
	/// @return												Number of opened Texture Views.
	//----------------------------------------------------------------------------------------
	Int32 GetViewCount();

	//----------------------------------------------------------------------------------------
	///	Get an array of unique Texture View IDs.
	/// @return												Array of unique Texture View IDs.
	//----------------------------------------------------------------------------------------
	maxon::BaseArray<Int32>* GetViewIds();

	//----------------------------------------------------------------------------------------
	///	Get the ID of the Texture View that currently has focus (if any).
	/// @return												ID of the Texture View which has focus, NOTOK if none found.
	//----------------------------------------------------------------------------------------
	Int32 GetActiveViewId();

	//----------------------------------------------------------------------------------------
	/// Remove focus from Texture View which currently has focus.
	/// @return												Returns true if successful.
	//----------------------------------------------------------------------------------------
	Bool ResetActiveView();

	//----------------------------------------------------------------------------------------
	/// Get the flag indicating if the Texture View in focus updates with selection changes.
	/// @return												Returns true if context updates are enabled, false otherwise.
	//----------------------------------------------------------------------------------------
	Bool GetAllowSelectUpdate();

	//----------------------------------------------------------------------------------------
	/// Set the flag indicating if the Texture View in focus updates with selection changes.
	/// @param[in] _allowSelectUpdate	Set to true to enable context updates.
	/// @return												Returns true if successful.
	//----------------------------------------------------------------------------------------
	Bool SetAllowSelectUpdate(Bool _allowSelectUpdate);
};

//----------------------------------------------------------------------------------------
///	A Texture View instance
//----------------------------------------------------------------------------------------
class PaintView
{
private:
	PaintView();
	~PaintView();

public:

	//----------------------------------------------------------------------------------------
	/// Free a PaintView class instance.
	/// @param[in] p									The PaintView instance.
	//----------------------------------------------------------------------------------------
	static void Free(PaintView *&p);

	//----------------------------------------------------------------------------------------
	/// Get the "UV Mesh > Show UV Mesh" option status for the view.
	/// @param[out] showMesh					True if enabled.
	/// @return												Returns true if successful.
	//----------------------------------------------------------------------------------------
	Bool GetShowMesh(Bool& showMesh);

	//----------------------------------------------------------------------------------------
	/// Set the "UV Mesh > Show UV Mesh" option status for the view.
	/// @param[in] showMesh						True to enable the option.
	/// @return												Returns true if successful.
	//----------------------------------------------------------------------------------------
	Bool SetShowMesh(Bool showMesh);

	//----------------------------------------------------------------------------------------
	/// Update the view context with the current selection.
	/// @return												Returns true if successful.
	//----------------------------------------------------------------------------------------
	Bool SetContextFromSelection();

	//----------------------------------------------------------------------------------------
	/// Apply the "View > Fit to Screen" feature.
	/// @return												Returns true if successful.
	//----------------------------------------------------------------------------------------
	Bool FitToScreen();
};

class iPaintManager;
class iPaintView;

PaintManager* GetPaintManager();
PaintView*		GetPaintView(PaintManager* paintManager, Int32 viewId);

//----------------------------------------------------------------------------------------
///	Paint Settings for BodyPaint Channels
//----------------------------------------------------------------------------------------
class BPPaintSettings
{
public:

private:
	BPPaintSettings();
	~BPPaintSettings();
};

//----------------------------------------------------------------------------------------
///	Color Settings for a BodyPaint Channel
//----------------------------------------------------------------------------------------
class BPSingleColorSettings : public BPPaintSettings
{
public:

	//----------------------------------------------------------------------------------------
	/// Set the RGB color and intensity value for the Color. The intensity value is internally multiplied against the RGB value.
	/// An RGB value of (0.5,0.5,0.5) is Grey.  If the RGB value is (1.0,1.0,1.0) and the intensity value is 0.5 then the resulting color is multipled agains this also resulting in Grey (0.5,0.5,0.5).
	/// @param[in] rgb								The RGB color.
	/// @param[in] i									The intensity value.
	//----------------------------------------------------------------------------------------
	void SetRGB(const Vector &rgb, Float i);

	//----------------------------------------------------------------------------------------
	/// Get the RGB color and intensity value for the Color. The intensity value is internally multiplied against the RGB value.
	/// An RGB value of (0.5,0.5,0.5) is Grey.  If the RGB value is (1.0,1.0,1.0) and the intensity value is 0.5 then the resulting color is multipled agains this also resulting in Grey (0.5,0.5,0.5).
	/// @param[out] rgb								The RGB color.
	/// @param[out] i									The intensity value.
	//----------------------------------------------------------------------------------------
	void GetRGB(Vector &rgb, Float &i);

private:
	BPSingleColorSettings();
	~BPSingleColorSettings();
};


//----------------------------------------------------------------------------------------
///	Color Settings
//----------------------------------------------------------------------------------------
class BPColorSettings : public BPPaintSettings
{
public:

	//----------------------------------------------------------------------------------------
	/// Get the BPSingleColorSettings for a specific channel.
	/// @param[in] channelid					The id of the channel to retrieve the color information for. IE CHANNEL_COLOR.
	/// @return												The color settings.
	//----------------------------------------------------------------------------------------
	BPSingleColorSettings *GetChannel(Int32 channelid);

private:
	BPColorSettings();
	~BPColorSettings();
};


//----------------------------------------------------------------------------------------
///	Functions to get access to the selected BodyPaint channel color settings
//----------------------------------------------------------------------------------------
struct BPColorSettingsHelpers
{
	//----------------------------------------------------------------------------------------
	///	Get the color settings, for the foreground or background color, of a BodyPaint Channel.
	/// @param[in] bg									Set to true to get the background color settings or false for the foreground color settings.
	/// @return												The color settings for the global background of foreground.
	//----------------------------------------------------------------------------------------
	static BPColorSettings *GetColorSettings(Bool bg);

	//----------------------------------------------------------------------------------------
	///	Get the color settings, for the foreground or background color, of a specific BodyPaint Channel.
	/// @param[in] bg									Set to true to get the background color settings or false for the foreground color settings.
	/// @param[in] channelid					The channel to get the color settings for. IE CHANNEL_COLOR.
	/// @return												The color settings.
	//----------------------------------------------------------------------------------------
	static BPSingleColorSettings *GetChannelColorSettings(Bool bg, Int32 channelid);

	//----------------------------------------------------------------------------------------
	///	Get the color settings for the currently selected channel in BodyPaint. If GetMultiMode is on and there are multiple channels selected then you can also specify the which channel.
	/// @param[in] bg									Set to true to get the background color settings or false for the foreground color settings. Special exception, set to NOTOK to use the selected background color.
	/// @param[in] channelid					The specific channel to get the color settings for, if GetMultiMode is enabled and there are multiple channels selected.	IE CHANNEL_COLOR.
	/// @return												The color settings.
	//----------------------------------------------------------------------------------------
	static BPSingleColorSettings *GetSelectedSingleColorSettings(Bool bg, Int32 *channelid = nullptr);

	//----------------------------------------------------------------------------------------
	///	Set which channel is selected.
	/// @param[in] bg									Set to true to get the background color settings or false for the foreground color settings. Special exception, set to NOTOK to use the selected background color.
	/// @param[in] channelid					The specific channel to enable. IE CHANNEL_COLOR.
	//----------------------------------------------------------------------------------------
	static void SetSelectedColorSettings(Bool bg, Int32 channelid);

	//----------------------------------------------------------------------------------------
	///	Get which channel is selected/enabled. If GetMultiMode is true then multiple channels could be enabled for painting.
	/// @param[in] channelid					The id of the channel. IE CHANNEL_COLOR.
	/// @return												Returns true if the channel is enabled. Returns false if not.
	//----------------------------------------------------------------------------------------
	static Bool GetGlobalChannelEnabling(Int32 channelid);

	//----------------------------------------------------------------------------------------
	///	Set which channel is selected/enabled. If GetMultiMode is true then you can set enable multiple channels for painting.
	/// @param[in] channelid					The id of the channel to enable. IE CHANNEL_COLOR.
	/// @param[in] enable							Set to true to enable the channel or false to disable it.
	//----------------------------------------------------------------------------------------
	static void SetGlobalChannelEnabling(Int32 channelid, Bool enable);

	//----------------------------------------------------------------------------------------
	///	If GetMultiMode is enabled then multiple channels can be selected/enabled in BodyPaint.
	/// @return												Gets the state of the MultiMode switch in BodyPaint.
	//----------------------------------------------------------------------------------------
	static Bool GetMultiMode();

	//----------------------------------------------------------------------------------------
	///	Set the state of the MultiMode button in BodyPaint. If GetMultiMode is enabled then multiple channels can be selected/enabled in BodyPaint.
	/// @param[in] multi							Set to true to enable or false to disable.
	//----------------------------------------------------------------------------------------
	static void SetMultiMode(Bool multi);

	//----------------------------------------------------------------------------------------
	/// Toggle the Foreground and Background colors.
	/// @return												Returns true if successful.
	//----------------------------------------------------------------------------------------
	static Bool ToggleFgBg(void);

	//----------------------------------------------------------------------------------------
	/// Get the active color preset.
	/// @return												The paint settings for the color preset.
	//----------------------------------------------------------------------------------------
	static BPPaintSettings *GetActiveColorPreset();

	//----------------------------------------------------------------------------------------
	/// !PRIVATE!
	//----------------------------------------------------------------------------------------
	static void SetActiveColorPreset(BPPaintSettings *set);

	//----------------------------------------------------------------------------------------
	/// !PRIVATE!
	//----------------------------------------------------------------------------------------
	static Bool SaveColorSettings(BPPaintSettings *set);

	//----------------------------------------------------------------------------------------
	/// !PRIVATE!
	//----------------------------------------------------------------------------------------
	static Bool DefineColorPreset(PaintTexture *tex, PaintLayerBmp *layer, PaintLayerBmp *mask, BPSingleColorSettings *color, Bool wallpaper);
};


struct PaintLibrary : public C4DLibrary
{
	PaintManager*							(*GetPaintManager)();
	PaintView*								(*GetPaintView)(PaintManager* paintManager, Int32 viewId);

	// iPaintManager methods
	Int32											(iPaintManager::*GetViewCount)();
	maxon::BaseArray<Int32>*	(iPaintManager::*GetViewIds)();
	Int32											(iPaintManager::*GetActiveViewId)();
	Bool											(iPaintManager::*ResetActiveView)();
	Bool											(iPaintManager::*GetAllowSelectUpdate)();
	Bool											(iPaintManager::*SetAllowSelectUpdate)(Bool _allowSelectUpdate);

	// iPaintView methods
	void											(*Free)(iPaintView *&p);
	Bool											(iPaintView::*GetShowMesh)(Bool &showMesh);
	Bool											(iPaintView::*SetShowMesh)(Bool showMesh);
	Bool											(iPaintView::*SetContextFromSelection)();
	Bool											(iPaintView::*FitToScreen)();



		//BPSingleColorSettings
	void	(*bpsinglecolorsettingsSetRGB)(BPSingleColorSettings *settings, const Vector &rgb, Float i);
	void	(*bpsinglecolorsettingsGetRGB)(BPSingleColorSettings *settings, Vector &v, Float &h);

	//BPColorSettings
	BPSingleColorSettings* (*bpcolorsettingsGetChannel)(BPColorSettings* settings, Int32 channelid);

	//BPColorSettingsHelpers
	BPColorSettings*				(*bpcolorsettingshelpersGetColorSettings)(Bool bg);
	BPSingleColorSettings*	(*bpcolorsettingshelpersGetChannelColorSettings)(Bool bg, Int32 channelid);
	BPSingleColorSettings*	(*bpcolorsettingshelpersGetSelectedSingleColorSettings)(Bool bg, Int32 *channelid);
	void										(*bpcolorsettingshelpersSetSelectedColorSettings)(Bool bg, Int32 channelid);
	Bool										(*bpcolorsettingshelpersGetGlobalChannelEnabling)(Int32 channelid);
	void										(*bpcolorsettingshelpersSetGlobalChannelEnabling)(Int32 channelid, Bool enable);
	Bool										(*bpcolorsettingshelpersGetMultiMode)(void);
	void										(*bpcolorsettingshelpersSetMultiMode)(Bool multi);
	Bool										(*bpcolorsettingshelpersToggleFgBg)(void);
	BPPaintSettings*				(*bpcolorsettingshelpersGetActiveColorPreset)();
	void										(*bpcolorsettingshelpersSetActiveColorPreset)(BPPaintSettings *set);
	Bool										(*bpcolorsettingshelpersSaveColorSettings)(BPPaintSettings *set);
	Bool										(*bpcolorsettingshelpersDefineColorPreset)(PaintTexture *tex, PaintLayerBmp *layer, PaintLayerBmp *mask, BPSingleColorSettings *color, Bool wallpaper);

};

#endif // LIB_PAINT_H__
