/////////////////////////////////////////////////////////////
// CINEMA 4D SDK                                           //
/////////////////////////////////////////////////////////////
// (c) MAXON Computer GmbH, all rights reserved            //
/////////////////////////////////////////////////////////////

#ifndef C4D_MEMORY_H__
#define C4D_MEMORY_H__

#ifdef __API_INTERN__
	#include "ge_memory.h"
#else

#include <new>
#ifdef MAXON_TARGET_OSX
	using namespace std;
#elif !defined MAXON_TARGET_WINDOWS
	#include <stddef.h>
#endif

#include <string.h>
#include <stdlib.h>
#include "c4d_plugin.h"
#include "operatingsystem.h"

namespace maxon
{
	void* GeNewMem(std::size_t s, int line, const char* file, bool clear);	// noexcept
	void GeDeleteMem(void* p);																							// noexcept
}


void _GeCheck(void* memptr);	// only for debug purposes
void _GeCheckAllMemory();			// only for debug purposes
void _GeFree(void** Daten);
Bool GeGetAllocSize(void* p, Int* out_size);
Bool IsAlienMem(void* p);

Bool GeGetAllocatorStatistics(BaseContainer& stat, void* allocator = nullptr);
#define	C4D_ALLOCATOR_STAT_CNT				1	// Int32
#define	C4D_ALLOCATOR_STAT_TOTAL_USED	2	// Int64
#define	C4D_ALLOCATOR_STAT_TOTAL_FREE	3	// Int64
#define	C4D_ALLOCATOR_INFO_CONTAINER	4	// BaseContainer
#define	C4D_ALLOCATOR_INFO_SLOT_SIZE	1	// Int32
#define	C4D_ALLOCATOR_INFO_SLOT_USED	2	// Int64
#define	C4D_ALLOCATOR_INFO_SLOT_FREE	3	// Int64

UInt GeMemGetFreePhysicalMemoryEstimate(void);

#ifdef __LEGACY_API
	#define GeAlloc(x)							 (C4DOS.Ge->Alloc((x), __LINE__, __FILE__))
	#define GeAllocNC(x)						 (C4DOS.Ge->AllocNC((x), __LINE__, __FILE__))
	#define GeReallocNC(p, s)				 (C4DOS.Ge->ReallocNC((p), (s), __LINE__, __FILE__))
	#define GeReallocTypeNC(t, p, s) (t*)(C4DOS.Ge->ReallocNC((p), sizeof(t) * (s), __LINE__, __FILE__))
	#define GeAllocType(t, x)				 (t*)(C4DOS.Ge->Alloc(sizeof(t) * (x), __LINE__, __FILE__))
	#define GeAllocTypeNC(t, x)			 (t*)(C4DOS.Ge->AllocNC(sizeof(t) * (x), __LINE__, __FILE__))
	#define GeFree(x)								 _GeFree((void**)(&(x)))
#endif

#define _GeAlloc(s, l, f)				 (C4DOS.Ge->Alloc((s), (l), (f)))
#define _GeAllocNC(s, l, f)			 (C4DOS.Ge->AllocNC((s), (l), (f)))
#define _GeReallocNC(p, s, l, f) (C4DOS.Ge->ReallocNC((p), (s), (l), (f)))

#ifdef USE_API_MAXON
	#include "maxon/defaultallocator.h"
#else
	#include "c4d_misc/memory/defaultallocator.h"
#endif

#ifdef MAXON_TARGET_64BIT
	#define	__C4D_MEM_ALIGNMENT_MASK__ 15	// memory returned by NewMem() is guaranteed to be at least aligned to a 16 byte boundary
#else
	#define	__C4D_MEM_ALIGNMENT_MASK__ 7	// memory returned by NewMem() is guaranteed to be at least aligned to and 8 byte boundary
#endif

#ifndef C4D_ALIGN
	#if defined MAXON_TARGET_WINDOWS
		#define C4D_ALIGN(_x_, _a_)	__declspec(align(_a_)) _x_
	#else
		#define C4D_ALIGN(_x_, _a_)	_x_ __attribute__((aligned(_a_)))
	#endif
#endif

inline void ClearMem(void* d, Int size, Int32 value = 0)
{
	memset(d, value & 0xFF, size);
}

inline void CopyMem(const void* s, void* d, Int size)
{
	if (s && d && size > 0)
		memcpy(d, s, size);
	else
		DebugStop();
}

inline void MemCopy(void* d, const void* s, Int size)
{
	if (s && d && size > 0)
		memcpy(d, s, size);
	else
		DebugStop();
}

template <class T, class U> inline void	FillMemTypeTemplate(T* data_ptr, Int size, const U* check_type, Int32 value)
{
#ifdef MAXON_TARGET_DEBUG
	if (check_type != data_ptr)	// this will cause a compile time error if the types are different
		return;
#endif

	memset(data_ptr, value, size);
}

#define	FillMemType(t, d, x, v) FillMemTypeTemplate(d, sizeof(t) * (x), ((t*) d), v)

//----------------------------------------------------------------------------------------
/// Clears memory of a datatype with pattern, if size >0.
/// @param[out] data_ptr					Address of the datatype.
/// @param[in] cnt								Number of elements t be filled with pattern (>1 e.g. for arrays), can be 0.
/// THREADSAFE
//----------------------------------------------------------------------------------------
template <typename T> inline void ClearMemType(T* data_ptr, Int cnt)
{
	memset(data_ptr, 0, size_t(cnt * sizeof(T)));
}

//----------------------------------------------------------------------------------------
/// Copies the content of a datatype to another of the same kind
/// source and destination memory must not overlap.
/// @param[in] src_ptr						Source address of datatype.
/// @param[out] dst_ptr						Destination address of datatype.
/// @param[in] cnt								Number of elements to be copied (>1 e.g. for arrays), can be 0.
/// THREADSAFE
//----------------------------------------------------------------------------------------
template <typename T> inline void CopyMemType(const T* src_ptr, T* dst_ptr, Int cnt)
{
	memcpy(dst_ptr, src_ptr, size_t(cnt * sizeof(T)));
}

// Please use NewObj, DeleteObj, NewMem, DeleteMem and BaseArray<> instead of
// gNew, gDelete, GeAlloc, GeFree and bNewDeprecatedUseArraysInstead/bDelete.
// The latter are deprecated and will be removed. DO NOT USE THIS - if your
// plugin still requires this: REFACTOR IT!
#ifdef __LEGACY_API
	inline void* operator	new(std::size_t s)
	{
		return maxon::GeNewMem(s, 0, nullptr, true);
	}

	inline void* operator	new[](std::size_t s)
	{
		return maxon::GeNewMem(s, 0, nullptr, true);
	}

	inline void* operator	new(std::size_t s, const std::nothrow_t&) throw()
	{
		return maxon::GeNewMem(s, 0, nullptr, true);
	}

	inline void* operator	new[](std::size_t s, const std::nothrow_t&) throw()
	{
		return maxon::GeNewMem(s, 0, nullptr, true);
	}

	inline void	operator	delete(void* p) throw()
	{
		maxon::GeDeleteMem(p);
	}

	inline void	operator	delete[](void* p) throw()
	{
		maxon::GeDeleteMem(p);
	}

	inline void	operator	delete(void* p, const std::nothrow_t&) throw()
	{
		maxon::GeDeleteMem(p);
	}

	inline void	operator	delete[](void* p, const std::nothrow_t&) throw()
	{
		maxon::GeDeleteMem(p);
	}

	// these are c4d's new/delete operators that don't replace the default operators
	inline void* operator	new(std::size_t s, const std::nothrow_t&, int line, const Char* file) throw()
	{
		return maxon::GeNewMem(s, line, file, true);
	}

	inline void* operator	new[](std::size_t s, const std::nothrow_t&, int line, const Char* file) throw()
	{
		return maxon::GeNewMem(s, line, file, true);
	}

	inline void	operator	delete(void* p, const std::nothrow_t&, int line, const Char* file) throw()
	{
		maxon::GeDeleteMem(p);
	}

	inline void	operator	delete[](void* p, const std::nothrow_t&, int line, const Char* file) throw()
	{
		maxon::GeDeleteMem(p);
	}


	// template for classes with a private delete operator
	#define	gDelete_TEMPLATE		\
	template <typename T> static void gDelete(T * &v)	\
	{	\
		if (v)	\
		{	\
			v->~T();	\
			operator delete(v, std::nothrow, 0, nullptr);	\
			v = nullptr;	\
		}	\
	}

	// Instead of replacing the delete operator use a template which calls the destructor manually and frees the memory
	template <typename T> void gDelete(T*& v)
	{
		if (v)
		{
			v->~T();
			C4DOS.Ge->Free(v);	// operator delete( v, std::nothrow, 0, nullptr );
			v = nullptr;
		}
	}

	// Unfortunately this requires intercepting operator delete
	#define bDelete(v) { delete[] v; v = nullptr; }

	///
	#define gNew new(std::nothrow, __LINE__, __FILE__)
	#define bNew new(std::nothrow, __LINE__, __FILE__)

#else
	#define	gDelete_TEMPLATE

	/// Substitute for bNew T[cnt]. This does not require us to intercept operator new[]/delete[], but
	/// it is more cumbersome, no faster and less secure than using a BaseArray. It is deprecated and will
	/// be removed.
	template <typename T> inline T* bNewDeprecatedUseArraysInstead(Int cnt)
	{
		void* mem = _GeAlloc((__C4D_MEM_ALIGNMENT_MASK__ + 1) + (sizeof(T) * cnt), __LINE__, __FILE__);
		if (mem)
		{
			*(Int*) mem = cnt;

			T* data =  (T*) (UInt(mem) + (__C4D_MEM_ALIGNMENT_MASK__ + 1));
			mem = (void*) data;

			for (Int i = 0; i < cnt; i++)
				new (&data[i]) T;
		}
		return (T*) mem;
	}

	/// Deprecated. Use BaseArray which is as fast, but easier to use and more secure.
	template <typename T> inline void bDelete(T*& v)
	{
		if (v)
		{
			void* data = (void*) (UInt(v) - (__C4D_MEM_ALIGNMENT_MASK__ + 1));
			Int cnt =* (Int*) data;

			for (Int i = 0; i < cnt; i++)
				v[i].~T();

			_GeFree((void**)&data);
			v = nullptr;
		}
	}

#endif


#include "ge_autoptr.h"

#define MEMORYPOOL_DEFAULT_BLOCKSIZE 0x80000

class DeprecatedMemoryPool
{
private:
	DeprecatedMemoryPool();
	~DeprecatedMemoryPool();

public:
	static DeprecatedMemoryPool* Alloc(Int block_size) { return C4DOS.Pl->Alloc(block_size); }
	static DeprecatedMemoryPool* Alloc() { return C4DOS.Pl->Alloc(MEMORYPOOL_DEFAULT_BLOCKSIZE); }
	static void Free(DeprecatedMemoryPool*& pool)
	{
		if (pool)
			C4DOS.Pl->Free(pool);
		pool = nullptr;
	}

	void* MemAlloc(Int size, Bool clear = true) { return C4DOS.Pl->AllocElement(this, size, clear); }
	void MemFree(void* mem, Int size) { C4DOS.Pl->FreeElement(this, mem, size); }
	void* MemAllocS(Int size, Bool clear = true) { return C4DOS.Pl->AllocElementS(this, size, clear); }
	void MemFreeS(void* mem) { C4DOS.Pl->FreeElementS(this, mem); }
	void* MemReAllocS(void* old, Int size, Bool clear = true) { return C4DOS.Pl->ReAllocElementS(this, old, size, clear); }
};

#endif

#endif // C4D_MEMORY_H__
