/////////////////////////////////////////////////////////////
// CINEMA 4D SDK                                           //
/////////////////////////////////////////////////////////////
// (c) MAXON Computer GmbH, all rights reserved            //
/////////////////////////////////////////////////////////////

#ifndef C4D_NETWORK_H__
#define C4D_NETWORK_H__

#ifndef __API_INTERN__
	#include "c4d_string.h"
	#include "c4d_basecontainer.h"
	#include "ge_prepass.h"
#else
	#include "ge_string.h"
#endif

#include "ge_math.h"
#ifdef USE_API_MAXON
	#include "maxon/basearray.h"
#else
	#include "c4d_misc/datastructures/basearray.h"
#endif

class NetworkIpAddrPort;
class ZeroConfTask;
class NetworkIpConnection;

#ifndef MAXON_TARGET_WINDOWS
	#define INVALID_SOCKET -1
#endif

#define NETWORK_INTERFACE_ANY	0

#ifndef AF_INET
	#define AF_INET 2
#endif

#ifndef AF_INET6
	#define AF_INET6 30
#endif

//--Common Network API--

#define NetCall(fnc) (*C4DOS.Ne->fnc)
#define IpCall(fnc)	 (this->*C4DOS.Ne->fnc)

class NetworkIpAddr
{
protected:
	NetworkIpAddr(_DONTCONSTRUCT v);

public:
	NetworkIpAddr();
	NetworkIpAddr(const NetworkIpAddr& a);
	NetworkIpAddr(UChar a, UChar b, UChar c, UChar d);
	NetworkIpAddr(Int16 x1, Int16 x2, Int16 x3, Int16 x4, Int16 x5, Int16 x6, Int16 x7, Int16 x8);
	~NetworkIpAddr();

	Bool SetIPv4(UChar a, UChar b, UChar c, UChar d);
	Bool SetIPv6(Int16 x1, Int16 x2, Int16 x3, Int16 x4, Int16 x5, Int16 x6, Int16 x7, Int16 x8);

	Bool GetIPv4(UChar& a, UChar& b, UChar& c, UChar& d) const;
	Bool GetIPv6(Int16& x1, Int16& x2, Int16& x3, Int16& x4, Int16& x5, Int16& x6, Int16& x7, Int16& x8) const;

	Bool Compare(const NetworkIpAddr& adr) const;

	Bool Write(HyperFile* hf) const;
	Bool Read(HyperFile* hf);

	Bool CopyTo(NetworkIpAddr& dst) const;

	Bool IsPrivateAddress() const;

	Bool IsValid() const;
	Bool IsEmpty() const;

	String		GetString(Int32 port = 0) const;
	PROTOCOL	GetProtocol() const;

	void Flush();

	const NetworkIpAddr&		operator =(const NetworkIpAddr& source);
	const NetworkIpAddrPort operator +(Int32 port) const;

	Bool operator ==(const NetworkIpAddr& adr) const
	{
		return Compare(adr);
	}

	Bool operator !=(const NetworkIpAddr& adr) const
	{
		return !Compare(adr);
	}

private:
	C4D_RESERVE_PRIVATE_TYPE(Int, dummyForRef);
};

class NetworkIpAddrPort : public NetworkIpAddr
{
	typedef NetworkIpAddr SUPER;

public:
	NetworkIpAddrPort();
	NetworkIpAddrPort(const NetworkIpAddr& a, Int32 port);
	NetworkIpAddrPort(UChar a, UChar b, UChar c, UChar d, Int32 port);
	NetworkIpAddrPort(Int16 x1, Int16 x2, Int16 x3, Int16 x4, Int16 x5, Int16 x6, Int16 x7, Int16 x8, Int32 port);
	~NetworkIpAddrPort();

	void Flush();

	Bool Write(HyperFile* hf) const;
	Bool Read(HyperFile* hf);

	Bool CopyTo(NetworkIpAddrPort& dst) const;

	String GetString(Bool getPort = true) const;

	Bool IsValid() const;

	void SetPort(Int32 port);
	Int32 GetPort() const;

	const NetworkIpAddrPort& operator =(const NetworkIpAddrPort& source);

	Bool operator ==(const NetworkIpAddrPort& a) const;
	Bool operator !=(const NetworkIpAddrPort& a) const;

private:
};

class NetworkInterface
{
public:
	void SetInterfaceIndex(Int interfaceIndex);
	Int GetInterfaceIndex() const;

	void SetInterfaceName(const String& interfaceName);
	String GetInterfaceName() const;

	void GetMacAddress(maxon::BaseArray<UChar>& macAddress) const;

	String GetDescription() const;
	void SetDescription(const String& description);

	Bool IsLoopback() const;
	void SetLoopback(Bool isLoopback);

	Int GetCountAddress() const;

	void Flush();

	Bool Append(const NetworkIpAddr& ipAddr, const NetworkIpAddr& broadcastAddress, const NetworkIpAddr& subnetMask);

	NetworkIpAddr GetIpAddress(Int i) const;
	NetworkIpAddr GetBroadcastAddress(Int i) const;
	NetworkIpAddr GetSubnetMask(Int i) const;

	Bool Write(HyperFile* hf) const;
	Bool Read(HyperFile* hf);
	Bool CopyTo(NetworkInterface& networkInterface) const;

	BaseContainer GetCustomData() const;
	void SetCustomData(const BaseContainer& customData);

	NetworkInterface* GetClone() const;

	static NetworkInterface* Alloc() { return NetCall(NetworkInterfaceAlloc) (); }
	static void Free(NetworkInterface*& p) { NetCall(NetworkInterfaceFree) (p); }

private:
	NetworkInterface();
	~NetworkInterface();
};

class ZeroConfService
{
	ZeroConfService() {}
	~ZeroConfService() {}

public:
	Bool SetTXTRecord(const String* keys, const String* values, Int cnt);
	Bool RemoveTXTRecord();
};

class NetworkIpConnection
{
	NetworkIpConnection() {}
	~NetworkIpConnection() {}
public:
	NetworkIpAddrPort GetRemoteAddr() const;
	NetworkIpAddrPort GetHostAddr() const;

	Int64 GetTransferedBytes() const;
};


//----------------------------------------------------------------------------------------
/// Wakes a computer up by WOL (Wake-On-LAN). Currently only IPv4 is supported. To get
/// WOL to work, a broadcast message on port '9' will be sent.
///
/// @param[in] macAddress					Mac address of the target machine.
/// @param[in] broadcast					Broadcast address where the message should be send to.
/// @return												True on success if the message was sent, otherwise false.
//----------------------------------------------------------------------------------------
Bool WakeOnLan(const maxon::BaseArray<UChar>& macAddress, const NetworkIpAddr& broadcast);

//----------------------------------------------------------------------------------------
/// Get the mac address of the machine. When this function returns false an error occurred, but then macAddress is filled at least with a
/// random-generated macAddress. The random-generated mac address will be changed after a restart of the host application.
/// @param[in] macAddress					Filled with the mac address.
/// @return												True on success, otherwise false.
//----------------------------------------------------------------------------------------
Bool GetMacAddress(maxon::BaseArray<UChar>& macAddress);

//----------------------------------------------------------------------------------------
/// Returns the best network interface to the specified protocol.
/// @param[in] protocol						Specified protocol.
/// @return												The found network interface.
//----------------------------------------------------------------------------------------
NetworkInterface* GetBestNetworkInterface(PROTOCOL protocol = PROTOCOL_IPV4);

//----------------------------------------------------------------------------------------
/// Returns all current active network interface objects.
/// @param[out] networkInterfaces	Assigned all network interfaces.
//----------------------------------------------------------------------------------------
Bool GetAllNetworkInterfaces(maxon::BaseArray<NetworkInterface*>& networkInterfaces);

//----------------------------------------------------------------------------------------
/// Returns the host name of the local machine.
/// @return												The name.
//----------------------------------------------------------------------------------------
String GetHostname();

//----------------------------------------------------------------------------------------
/// Returns the host name of the given ip address.
/// @return												The host name. Can be empty if the address could not be resolved.
//----------------------------------------------------------------------------------------
String GetHostname(const NetworkIpAddr& ipAddr);

//----------------------------------------------------------------------------------------
/// Returns the interface index. The interface name "all" returns 0.
/// @param[in] interfaceName			The name.
/// @return												The index.
//----------------------------------------------------------------------------------------
Int GetInterfaceIndex(const String& interfaceName);

//----------------------------------------------------------------------------------------
/// Resolves a host name. If the address object can be an ip address object, you should use GetIpAddress(address, a, true) instead.
/// @param[in] address						Host name to resolve.
/// @param[in] ptype							Pass PROTOCOL_IPV4 if the result should just contain IPv4 address, same for PROTOCOL_IPV6 with IPv6 address and PROTOCOL_ALL for all addresses.
/// @param[out] addrs							Contains the addresses of the host name when the function succeeded.
/// @param[in] firstMatch					Address array has a maximum size of 1 element.
/// @param[in] forceResolve				If false use the DNS cache if the address is a host name and already known. Use true for an automatic mode.
/// @return												Returns RESOLVERESULT_OK on success, otherwise the resolve operation failed.
//----------------------------------------------------------------------------------------
RESOLVERESULT ResolveHostname(const String& address, PROTOCOL ptype, maxon::BaseArray<NetworkIpAddr>& addrs, Bool firstMatch = true, Bool forceResolve = false);

//----------------------------------------------------------------------------------------
/// Converts a host name, IPv4 or IPv6 address into an ip address object. A port value in the address will be ignored. When the resolve
/// option is enabled, the call blocks and waits for the DNS system when the address is a host name. E.g:
///			address: Sebastians-Macbook-Pro.local (DNS system used, call blocks)
///			address: 192.168.10.1 (no DNS system used, call does not block)
///
/// @param[in] address						IP address or host name. If the string is a host name, resolve must be set to true otherwise the function fails.
/// @param[out] a									Assigned ip address object.
/// @param[in] resolve						Set to true to resolve the address via the DNS system.
/// @return												True on success, otherwise false.
//----------------------------------------------------------------------------------------
Bool GetIpAddress(const String& address, NetworkIpAddr& a, Bool resolve = false, Bool forceResolve = false);

//----------------------------------------------------------------------------------------
/// Converts a host name, IPv4 or IPv6 address into an ip address object. The address can have a port value. When the resolve option is enabled, the
/// call blocks and waits for the DNS system when the address is a host name. E.g:
///			address: Sebastians-Macbook-Pro.local:1234 (DNS system used, call blocks)
///			address: 192.168.10.1:1234 (no DNS system used, call does not block)
///
/// @param[in] address						IP address or host name. If the string is a host name, resolve must be set to true otherwise the function fails.
/// @param[out] a									Assigned IP address object.
/// @param[in] resolve						Set to true to resolve the address via the DNS system.
/// @return												True on success, otherwise false.
//----------------------------------------------------------------------------------------
Bool GetIpAddress(const String& address, NetworkIpAddrPort& a, Bool resolve = false, Bool forceResolve = false);

//----------------------------------------------------------------------------------------
/// Split a specified scheme with host name and port into it's components. Can handle IPv4, IPv6 addresses and host names.
/// For instance:
///
///		http://127.0.0.1:8080
///		http://[AB:AB:AB:AB:AB:AB]:1234
///		192.168.10.1
///		Sebastians-Macbook-Pro.local:80
///
/// @param[in] address						The host name to split. Can optionally contain a scheme, host and port.
/// @param[in] scheme							The assigned scheme, can be empty if no scheme was found.
/// @param[in] host								The assigned address/host name without port scheme and port number.
/// @param[in] port								The port, if no port was found port NOTOK is assigned.
//----------------------------------------------------------------------------------------
Bool SplitAddress(String address, String* scheme, String* host, Int32* port);

//----------------------------------------------------------------------------------------
/// Opens a connection. Supports IPv4.
/// @param[in] adr								Destination address to connect to.
/// @param[in] connectTimeout			Timeout in seconds for the initial connection.
/// @param[in] sessionTimeout			Timeout in seconds for receive and send operations.
/// @param[in] useNagleAlgorithm	Pass true to set the TCP/IP flag for TCP_NODELAY. For more information see: http://msdn.microsoft.com/en-us/library/ms817942.aspx.
/// @param[in] error							Assigned error if the connection failed.
/// @return												Connection object for the outgoing connection. Must be freed with CloseConnection.
//----------------------------------------------------------------------------------------
NetworkIpConnection* OpenOutgoing(const NetworkIpAddrPort& adr, BaseThread* thread = nullptr, Int connectTimeout = 30, Int sessionTimeout = 10, Bool useNagleAlgorithm = true, Int* error = nullptr);

//----------------------------------------------------------------------------------------
/// Opens a connection. Supports IPv4.
/// @param[in] adr								Destination address to connect to (ipv4 address or host name)
/// @param[in] connectTimeout			Timeout for the connect.
/// @param[in] sessionTimeout			Session timeout.
/// @param[in] useNagleAlgorithm	Pass true to set the TCP/IP flag for TCP_NODELAY. For more information see: http://msdn.microsoft.com/en-us/library/ms817942.aspx.
/// @param[in] error							Assigned error if the connection failed.
/// @return												Connection object for the outgoing connection. Must be freed with CloseConnection.
//----------------------------------------------------------------------------------------
NetworkIpConnection* OpenOutgoing(const String& adr, BaseThread* thread = nullptr, Int connectTimeout = 30, Int sessionTimeout = 10, Bool useNagleAlgorithm = true, Int* error = nullptr);

Int BytesInInputBuffer(NetworkIpConnection* ipc);
Int RecvBytes(NetworkIpConnection* ipc, void* buf, Int size);
Int SendBytes(NetworkIpConnection* ipc, const void* buf, Int size);
NetworkIpConnection* OpenListener(const NetworkIpAddrPort& adr, BaseThread* thread, Int sessionTimeout, Bool dontwait, Int* error);
NetworkIpConnection* OpenListener(const String& adr, BaseThread* thread, Int sessionTimeout, Bool useNagleAlgorithm, Int* error);
NetworkIpConnection* OpenWaitForIncoming(NetworkIpConnection* listener, BaseThread* connection, Int* error);
void KillConnection(NetworkIpConnection*& ipc);
void CloseConnection(NetworkIpConnection*& ipc);

//--ZeroConf API--

//----------------------------------------------------------------------------------------
/// Get the version number of the bonjour version.
/// @param[out] versionNumber			The main version number of bonjour.
/// @param[out] revisionNumber		The revision number.
/// @param[out] isRunning					If the function returns false, isRunning can be true to indicate that the bonjour service is running but this process cannot connect to it.
///																This can happen if the process is blocked by a firewall to connect to 127.0.0.1/localhost.
/// @return												True if Bonjour is installed, otherwise false.
//----------------------------------------------------------------------------------------
Bool GetBonjourVersion(Int& versionNumber, Int& revisionNumber, Bool& isRunning);

ZeroConfBrowser* StartZeroConfBrowser(const String& serviceType, ZeroConfBrowserCallback browseCallback, ZeroConfBrowserResolvedCallback resolveCallback, void* context);
void StopZeroConfBrowser(ZeroConfBrowser* browser);

ZeroConfService* RegisterZeroConfService(String serviceName, const String& serviceType, Int lInterface, Int32 port, const String& domainName = String());
void DeregisterZeroConfService(ZeroConfService* service);

#endif // C4D_NETWORK_H__
